//*****************************************************************************
//
//  Author:         Kenny Kerr
//  Date created:   6 August 2006
//
//  Description:    Implements the image window for the sample. The image 
//                  window is where most of the effects are applied.
//
//*****************************************************************************

#include "Precompiled.h"
#include "ImageWindow.h"
#include "DwmHelpers.h"

ImageWindow::ImageWindow() :
    m_enableComposition(true),
    m_enableBlur(true),
    m_extendFrame(false),
    m_disableTransitions(false),
    m_ncRenderingPolicy(DWMNCRP_USEWINDOWSTYLE),
    m_flip3DPolicy(DWMFLIP3D_DEFAULT),
    m_useLayeredWindow(false),
    m_margins(),
    m_isCompositionEnabled(false),
    m_allowNcPaint(false)
{
    m_margins.cxLeftWidth = -1;

    UpdateIsCompositionEnabled();
}

void ImageWindow::Create(const DWORD extendedStyles)
{
    VERIFY(__super::Create(0, 
                           0, 
                           L"DWM - Image Window", 
                           0, 
                           extendedStyles));

    EnableBlur();
    ExtendFrame();

    if (m_isCompositionEnabled)
    {
        Kerr::DwmSetWindowAttribute(m_hWnd, DWMWA_TRANSITIONS_FORCEDISABLED, m_disableTransitions);
        Kerr::DwmSetWindowAttribute(m_hWnd, DWMWA_NCRENDERING_POLICY, m_ncRenderingPolicy);
        Kerr::DwmSetWindowAttribute(m_hWnd, DWMWA_ALLOW_NCPAINT, m_allowNcPaint);
    }

    if (WS_EX_LAYERED == extendedStyles)
    {
        UpdateLayeredWindow();
    }
    else
    {
        if (m_image.IsNull())
        {
            VERIFY(ResizeClient(200,
                                200,
                                false));
        }
        else
        {
            VERIFY(ResizeClient(m_image.GetWidth(),
                                m_image.GetHeight(),
                                false));
        }
    }

    VERIFY(CenterWindow());
    ShowWindow(SW_SHOWNORMAL);
}

void ImageWindow::EnableBlur(const bool enable)
{
    if (enable != m_enableBlur)
    {
        m_enableBlur = enable;

        if (0 != m_hWnd)
        {
            EnableBlur();
        }
    }
}

void ImageWindow::DisableTransitions(const bool disable)
{
    if (disable != m_disableTransitions)
    {
        m_disableTransitions = disable;

        if (0 != m_hWnd)
        {
            Kerr::DwmSetWindowAttribute(m_hWnd, 
                                        DWMWA_TRANSITIONS_FORCEDISABLED,
                                        m_disableTransitions);
        }
    }
}

void ImageWindow::AllowNcPaint(const bool allow)
{
    if (allow != m_allowNcPaint)
    {
        m_allowNcPaint = allow;

        if (0 != m_hWnd)
        {
            Kerr::DwmSetWindowAttribute(m_hWnd, 
                                        DWMWA_ALLOW_NCPAINT,
                                        m_allowNcPaint);

            // Trick the window into reloading the image and window size.
            CString fileName = m_fileName;
            SetFileName(CString());
            SetFileName(fileName);
        }
    }
}

void ImageWindow::ExtendFrame(const bool extend)
{
    if (extend != m_extendFrame)
    {
        m_extendFrame = extend;

        if (0 != m_hWnd)
        {
            ExtendFrame();
        }
    }
}

void ImageWindow::UpdateExtendFrameMargins(const MARGINS& margins)
{
    m_margins = margins;

    if (m_extendFrame && 0 != m_hWnd)
    {
        ExtendFrame();
    }
}

void ImageWindow::UpdateNcRenderingPolicy(const DWMNCRENDERINGPOLICY policy)
{
    if (policy != m_ncRenderingPolicy)
    {
        m_ncRenderingPolicy = policy;

        if (0 != m_hWnd)
        {
            Kerr::DwmSetWindowAttribute(m_hWnd, 
                                        DWMWA_NCRENDERING_POLICY,
                                        m_ncRenderingPolicy);
        }
    }
}

void ImageWindow::UpdateFlip3DPolicy(const DWMFLIP3DWINDOWPOLICY policy)
{
    if (policy != m_flip3DPolicy)
    {
        m_flip3DPolicy = policy;

        if (0 != m_hWnd)
        {
            Kerr::DwmSetWindowAttribute(m_hWnd, 
                                        DWMWA_FLIP3D_POLICY,
                                        m_flip3DPolicy);
        }
    }
}

bool ImageWindow::SetFileName(const CString& fileName)
{
    bool updated = false;

    if (m_fileName != fileName)
    {
        std::auto_ptr<Gdiplus::Image> image(Gdiplus::Image::FromFile(fileName));

        if (Kerr::Succeeded(image))
        {
            CRect imageRect(0, 
                            0, 
                            image->GetWidth(), 
                            image->GetHeight());

            VERIFY(m_image.Create(imageRect.Width(),
                                  imageRect.Height(),
                                  32,
                                  CImage::createAlphaChannel));

            {
                Gdiplus::Graphics graphics(m_image.GetDC());
                GDIPLUS_VERIFY(graphics);

                if (!m_isCompositionEnabled)
                {
                    const COLORREF color = ::GetSysColor(COLOR_3DFACE);

                    GDIPLUS_VERIFY(graphics.Clear(Gdiplus::Color(GetRValue(color), 
                                                                 GetGValue(color),
                                                                 GetBValue(color))));
                }

                GDIPLUS_VERIFY(graphics.DrawImage(image.get(), 
                                                  0, 
                                                  0, 
                                                  imageRect.right, 
                                                  imageRect.bottom));
            }

            m_image.ReleaseDC();
            m_fileName = fileName;
            updated = true;
        }
        else if (!m_image.IsNull())
        {
            m_image.Destroy();
            m_fileName.Empty();
            updated = true;
        }

        if (0 != m_hWnd)
        {
            if (m_useLayeredWindow)
            {
                UpdateLayeredWindow();
            }
            else
            {
                if (!m_image.IsNull())
                {
                    if (m_allowNcPaint)
                    {
                        VERIFY(SetWindowPos(0, // relative window
                                            0, 0, // location
                                            m_image.GetWidth(),
                                            m_image.GetHeight(),
                                            SWP_NOMOVE | SWP_NOZORDER));
                    }
                    else
                    {
                        VERIFY(ResizeClient(m_image.GetWidth(),
                                            m_image.GetHeight(),
                                            false));
                    }
                }
                    
                VERIFY(Invalidate());
            }
        }
    }

    return updated;
}

void ImageWindow::CreateLayeredWindow(bool enable)
{
    if (m_useLayeredWindow != enable)
    {
        m_useLayeredWindow = enable;

        if (0 != m_hWnd)
        {
            VERIFY(DestroyWindow());

            if (m_useLayeredWindow)
            {
                Create(WS_EX_LAYERED);
            }
            else
            {
                Create();
                ShowWindow(SW_SHOWNORMAL);
            }
        }
    }
}

LRESULT ImageWindow::OnDwmCompositionEnabled(UINT /*message*/, 
                                             WPARAM /*wParam*/, 
                                             LPARAM /*lParam*/, 
                                             BOOL& /*handled*/)
{
    UpdateIsCompositionEnabled();

    // Trick the window into reloading the image and window size.
    CString fileName = m_fileName;
    SetFileName(CString());
    SetFileName(fileName);

    EnableBlur();
    ExtendFrame();
    return 0;
}

LRESULT ImageWindow::OnPaint(UINT /*message*/, 
                             WPARAM /*wParam*/, 
                             LPARAM /*lParam*/, 
                             BOOL& /*handled*/)
{
    if (!m_useLayeredWindow)
    {
        CPaintDC dc(m_hWnd);

        CRect clientRect;
        VERIFY(GetClientRect(&clientRect));

        if (m_image.IsNull())
        {
            CBrush brush;

            if (m_isCompositionEnabled)
            {
                VERIFY(brush.CreateSolidBrush(0));
            }
            else
            {
                VERIFY(brush.CreateSolidBrush(::GetSysColor(COLOR_3DFACE)));
            }

            VERIFY(dc.FillRect(&clientRect,
                               brush));
        }
        else if (m_allowNcPaint)
        {
            VERIFY(m_image.BitBlt(CWindowDC(m_hWnd), 0, 0));
        }
        else
        {
            VERIFY(m_image.BitBlt(dc, 0, 0));
        }
    }

    return 0;
}

bool ImageWindow::OnEraseBackground(HDC /*dc*/)
{
    return true;
}

LRESULT ImageWindow::OnNcHitTest(UINT message, 
                                 WPARAM wParam, 
                                 LPARAM lParam, 
                                 BOOL& /*handled*/)
{
    LRESULT result = DefWindowProc(message, 
                                   wParam, 
                                   lParam);

    if (HTCLIENT == result)
    {
        result = HTCAPTION;
    }

    return result;
}

void ImageWindow::UpdateLayeredWindow()
{
    if (!m_image.IsNull())
    {
        CPoint windowPosition(0, 0);
        CPoint layerPosition(0, 0);

        CSize size(m_image.GetWidth(),
                   m_image.GetHeight());

        Kerr::BlendFunction blendFunction;

        VERIFY(::UpdateLayeredWindow(m_hWnd,
                                     0, // obtain screen DC
                                     0,
                                     &size,
                                     m_image.GetDC(),
                                     &layerPosition,
                                     0, // no color key
                                     &blendFunction,
                                     ULW_ALPHA));

        m_image.ReleaseDC();

        ShowWindow(SW_SHOW);
    }
}

void ImageWindow::EnableBlur()
{
    ASSERT(0 != m_hWnd);

    if (m_isCompositionEnabled)
    {
        COM_VERIFY(Kerr::EnableBlurBehindWindow(m_hWnd,
                                                m_enableBlur));
    }
}

void ImageWindow::ExtendFrame()
{
    ASSERT(0 != m_hWnd);

    if (m_isCompositionEnabled)
    {
        const MARGINS dontExtend = { 0 };

        if (m_extendFrame)
        {
            COM_VERIFY(::DwmExtendFrameIntoClientArea(m_hWnd,
                                                      &m_margins));
        }
        else
        {
            COM_VERIFY(::DwmExtendFrameIntoClientArea(m_hWnd,
                                                      &dontExtend));
        }
    }
}

void ImageWindow::UpdateIsCompositionEnabled()
{
    BOOL legacyBool = FALSE;
    COM_VERIFY(::DwmIsCompositionEnabled(&legacyBool));
    m_isCompositionEnabled = FALSE != legacyBool;
}
