//*****************************************************************************
//
//  Author:         Kenny Kerr
//  Date created:   6 August 2006
//
//  Description:    Implements the main window for the sample. The main 
//                  window provides the "control panel" for the sample.
//
//*****************************************************************************

#include "Precompiled.h"
#include "MainWindow.h"

MainWindow::MainWindow() :
    m_isCompositionEnabled(false),
    m_compositionOpaqueBlend(false)
{
    VERIFY(Create(0));

    CMessageLoop* loop = _Module.GetMessageLoop();
    VERIFY(loop->AddMessageFilter(this));
    VERIFY(loop->AddIdleHandler(this));

    PrepareControls();
}

BOOL MainWindow::OnIdle()
{
    CButton showThumbnail(GetDlgItem(IDC_SHOW_THUMBNAIL));
    showThumbnail.EnableWindow(0 != m_imageWindow.m_hWnd);
    
    return FALSE;
}

void MainWindow::OnFinalMessage(HWND /*window*/)
{
    CMessageLoop* loop = _Module.GetMessageLoop();
    VERIFY(loop->RemoveMessageFilter(this));
    VERIFY(loop->RemoveIdleHandler(this));
}

BOOL MainWindow::PreTranslateMessage(MSG* message)
{
    m_toolTips.RelayEvent(message);

    return ::IsDialogMessage(m_hWnd, 
                             message);
}

HBRUSH MainWindow::OnCtlColorStatic(CDCHandle dc,
                                    HWND window)
{
    // This is used to paint the background of the status area.

    bool statusControls = m_status == window ||
                          m_statusBackground == window;

    HBRUSH brush = 0;

    if (statusControls)
    {
        dc.SetBkColor(::GetSysColor(COLOR_INFOBK));
        brush = static_cast<HBRUSH>(LongToPtr(COLOR_INFOBK + 1));
    }
    else
    {
        SetMsgHandled(false);
    }

    return brush;
}

LRESULT MainWindow::OnCompositionEnabled(UINT /*message*/, 
                                         WPARAM /*wParam*/, 
                                         LPARAM /*lParam*/, 
                                         BOOL& /*handled*/)
{
    UpdateIsCompositionEnabled();
    UpdateStatus();
    return 0;
}

LRESULT MainWindow::OnCompositionColorChanged(UINT /*message*/, 
                                              WPARAM wParam, 
                                              LPARAM lParam, 
                                              BOOL& /*handled*/)
{
    m_compositionColor.SetValue(static_cast<Gdiplus::ARGB>(wParam));
    m_compositionOpaqueBlend = 0 != lParam;

    UpdateStatus();
    return 0;
}

LRESULT MainWindow::OnDisableComposition(WORD /*notifyCode*/, 
                                         WORD /*identifier*/, 
                                         CButton checkBox, 
                                         BOOL& /*handled*/)
{
    const UINT action = BST_CHECKED == checkBox.GetCheck() ? DWM_EC_DISABLECOMPOSITION : DWM_EC_ENABLECOMPOSITION;
    COM_VERIFY(::DwmEnableComposition(action));

    return 0;
}

LRESULT MainWindow::OnEnableBlur(WORD /*notifyCode*/, 
                                 WORD /*identifier*/, 
                                 CButton checkBox, 
                                 BOOL& /*handled*/)
{
    m_imageWindow.EnableBlur(BST_CHECKED == checkBox.GetCheck());
    return 0;
}

LRESULT MainWindow::OnDisableTransitions(WORD /*notifyCode*/, 
                                         WORD /*identifier*/, 
                                         CButton checkBox, 
                                         BOOL& /*handled*/)
{
    m_imageWindow.DisableTransitions(BST_CHECKED == checkBox.GetCheck());
    return 0;
}

LRESULT MainWindow::OnAllowNcPaint(WORD /*notifyCode*/, 
                                   WORD /*identifier*/, 
                                   CButton checkBox, 
                                   BOOL& /*handled*/)
{
    m_imageWindow.AllowNcPaint(BST_CHECKED == checkBox.GetCheck());
    return 0;
}

LRESULT MainWindow::OnExtendFrame(WORD /*notifyCode*/, 
                                  WORD /*identifier*/, 
                                  CButton checkBox, 
                                  BOOL& /*handled*/)
{
    m_imageWindow.ExtendFrame(BST_CHECKED == checkBox.GetCheck());
    return 0;
}

LRESULT MainWindow::OnExtendFrameMargins(WORD /*notifyCode*/, 
                                         WORD /*identifier*/, 
                                         HWND /*window*/, 
                                         BOOL& /*handled*/)
{
    CString text;
    MARGINS margins = { -1 };

    if (0 < GetDlgItemText(IDC_EXTEND_LEFT, text))
    {
        margins.cxLeftWidth = _wtoi(text);
    }

    if (0 < GetDlgItemText(IDC_EXTEND_RIGHT, text))
    {
        margins.cxRightWidth = _wtoi(text);
    }

    if (0 < GetDlgItemText(IDC_EXTEND_TOP, text))
    {
        margins.cyTopHeight = _wtoi(text);
    }

    if (0 < GetDlgItemText(IDC_EXTEND_BOTTOM, text))
    {
        margins.cyBottomHeight = _wtoi(text);
    }

    m_imageWindow.UpdateExtendFrameMargins(margins);
    return 0;
}

LRESULT MainWindow::OnNcRenderingPolicy(WORD /*notifyCode*/, 
                                        WORD /*identifier*/, 
                                        CComboBox comboBox, 
                                        BOOL& /*handled*/)
{
    m_imageWindow.UpdateNcRenderingPolicy(static_cast<DWMNCRENDERINGPOLICY>(comboBox.GetCurSel()));
    return 0;
}

LRESULT MainWindow::OnFlip3DPolicy(WORD /*notifyCode*/, 
                                   WORD /*identifier*/, 
                                   CComboBox comboBox, 
                                   BOOL& /*handled*/)
{
    m_imageWindow.UpdateFlip3DPolicy(static_cast<DWMFLIP3DWINDOWPOLICY>(comboBox.GetCurSel()));
    return 0;
}

LRESULT MainWindow::OnFileName(WORD /*notifyCode*/, 
                               WORD /*identifier*/, 
                               CEdit edit, 
                               BOOL& /*handled*/)
{
    CString text;
    edit.GetWindowText(text);

    if (m_imageWindow.SetFileName(text) &&
        0 != m_thumbnailWindow.m_hWnd)
    {
        m_thumbnailWindow.Update();
    }

    return 0;
}

LRESULT MainWindow::OnCreateLayeredWindow(WORD /*notifyCode*/, 
                                          WORD /*identifier*/, 
                                          CButton checkBox, 
                                          BOOL& /*handled*/)
{
    const bool recreateThumbnailWindow = 0 != m_thumbnailWindow.m_hWnd;
    CRect thumbnailWindowRect;

    if (recreateThumbnailWindow)
    {
        VERIFY(m_thumbnailWindow.GetWindowRect(&thumbnailWindowRect));
        m_thumbnailWindow.SendMessage(WM_CLOSE);
    }

    m_imageWindow.CreateLayeredWindow(BST_CHECKED == checkBox.GetCheck());

    if (recreateThumbnailWindow)
    {
        m_thumbnailWindow.Create(&m_imageWindow, 
                                 &thumbnailWindowRect);
    }

    return 0;
}

LRESULT MainWindow::OnThumbnailExcludeNcArea(WORD /*notifyCode*/, 
                                             WORD /*identifier*/, 
                                             CButton checkBox, 
                                             BOOL& /*handled*/)
{
    m_thumbnailWindow.ExcludeNcArea(BST_CHECKED == checkBox.GetCheck());
    return 0;
}

LRESULT MainWindow::OnCancel(WORD /*notifyCode*/, 
                             WORD identifier, 
                             HWND /*window*/, 
                             BOOL& /*handled*/)
{
    if (0 != m_imageWindow.m_hWnd)
    {
        m_imageWindow.SendMessage(WM_CLOSE);
    }

    if (0 != m_thumbnailWindow.m_hWnd)
    {
        m_thumbnailWindow.SendMessage(WM_CLOSE);
    }

    VERIFY(DestroyWindow());
    ::PostQuitMessage(identifier);
    return 0;
}

LRESULT MainWindow::OnShowImageWindow(WORD /*notifyCode*/, 
                                      WORD /*identifier*/, 
                                      HWND /*window*/, 
                                      BOOL& /*handled*/)
{
    if (0 == m_imageWindow.m_hWnd)
    {
        m_imageWindow.Create(IsDlgButtonChecked(IDC_LAYERED_WINDOW) ? WS_EX_LAYERED : 0);
    }

    m_imageWindow.SetFocus();
    return 0;
}

LRESULT MainWindow::OnShowThumbnail(WORD /*notifyCode*/, 
                                    WORD /*identifier*/, 
                                    HWND /*window*/, 
                                    BOOL& /*handled*/)
{
    if (0 == m_thumbnailWindow.m_hWnd)
    {
        m_thumbnailWindow.Create(&m_imageWindow);
    }

    m_thumbnailWindow.SetFocus();
    return 0;
}

LRESULT MainWindow::OnLink(int /*identifier*/,
                           LPNMHDR lParam,
                           BOOL& /*handled*/)
{
    ASSERT(0 != lParam);
    PNMLINK link = reinterpret_cast<PNMLINK>(lParam);

    ::ShellExecute(0,
                   0,
                   link->item.szUrl,
                   0,
                   0,
                   SW_SHOWNORMAL);

    return 0;
}

void MainWindow::PrepareControls()
{
    ASSERT(0 != m_hWnd);

    VERIFY(m_toolTips.Create(m_hWnd));

    // Set text cues and tool-tips
    SetHelpText(IDC_FILENAME, L"Enter PNG image file name");
    SetHelpText(IDC_EXTEND_LEFT, L"Left");
    SetHelpText(IDC_EXTEND_RIGHT, L"Right");
    SetHelpText(IDC_EXTEND_TOP, L"Top");
    SetHelpText(IDC_EXTEND_BOTTOM, L"Bottom");

    CComboBox renderingPolicy(GetDlgItem(IDC_NC_RENDERING_POLICY));
    renderingPolicy.AddString(L"Use window style");
    renderingPolicy.AddString(L"Disabled (window style ignored)");
    renderingPolicy.AddString(L"Enabled (window style ignored)");
    renderingPolicy.SetCurSel(0);

    CComboBox flip3dPolicy(GetDlgItem(IDC_FLIP3D_POLICY));
    flip3dPolicy.AddString(L"Default");
    flip3dPolicy.AddString(L"Exclude below");
    flip3dPolicy.AddString(L"Exclude above");
    flip3dPolicy.SetCurSel(0);

    VERIFY(CheckDlgButton(IDC_BLUR, 
                          BST_CHECKED));

    m_statusBackground.Attach(GetDlgItem(IDC_STATUS_BACKGROUND));
    m_status.Attach(GetDlgItem(IDC_STATUS));

    UpdateIsCompositionEnabled();
    UpdateColorizationInfo();
    UpdateStatus();
    CreateLinkControl();

    VERIFY(SetWindowPos(0, // z-order
                        10, 10, // location
                        0, 0, // size
                        SWP_NOSIZE | SWP_NOZORDER));
}

void MainWindow::UpdateIsCompositionEnabled()
{
    BOOL legacyBool = FALSE;
    COM_VERIFY(::DwmIsCompositionEnabled(&legacyBool));
    m_isCompositionEnabled = FALSE != legacyBool;
}

void MainWindow::UpdateColorizationInfo()
{
    Gdiplus::ARGB color = 0;
    BOOL legacyBool = FALSE;

    COM_VERIFY(::DwmGetColorizationColor(&color,
                                         &legacyBool));

    m_compositionOpaqueBlend = FALSE != legacyBool;
    m_compositionColor.SetValue(color);
}

void MainWindow::UpdateStatus()
{
    CString text;

    if (m_isCompositionEnabled)
    {
        text.Append(L"Desktop composition is enabled.\n");
    }
    else
    {
        text.Append(L"Desktop composition is not enabled.\n");
    }

    text.AppendFormat(L"Composition color: A:%d R:%d G:%d B:%d\n", 
                      m_compositionColor.GetA(),
                      m_compositionColor.GetR(),
                      m_compositionColor.GetG(),
                      m_compositionColor.GetB());

    if (m_compositionOpaqueBlend)
    {
        text.Append(L"The composition is opaque.");
    }
    else
    {
        text.Append(L"The composition is translucent.");
    }

    m_status.SetWindowText(text);
}

void MainWindow::SetHelpText(int identifier,
                             PCWSTR text)
{
    ASSERT(0 != text);

    CEdit control(GetDlgItem(identifier));
    VERIFY(control.SetCueBannerText(text));

    VERIFY(m_toolTips.AddTool(control,
                              text));
}

void MainWindow::CreateLinkControl()
{
    CRect rect;
    CString text;
    CWindow placeholder(GetDlgItem(IDC_LINK));

    VERIFY(placeholder.GetWindowRect(&rect));
    VERIFY(ScreenToClient(&rect));
    placeholder.GetWindowText(text);

    CLinkCtrl link;

    VERIFY(link.Create(m_hWnd,
                       rect,
                       text,
                       WS_CHILD | WS_VISIBLE));

    link.SetFont(placeholder.GetFont());
    VERIFY(placeholder.DestroyWindow());
}
